from decimal import Decimal

from eth_utils.conversions import to_int
from mm_eth import abi, erc20, rpc
from mm_std import Err, Ok, Result

from mm_balance.config import Config
from mm_balance.types import EthTokenAddress, Network


def get_balance(address: str, token_address: str | None, config: Config) -> Result[Decimal]:
    if token_address is not None and token_address.lower() == EthTokenAddress.SDAI:
        return _get_sdai_balance(config.nodes[Network.ETH], address, config.proxies, config.round_ndigits, attempts=5)
    if token_address is not None:
        return erc20.get_balance(
            config.nodes[Network.ETH],
            token_address,
            address,
            proxies=config.proxies,
            attempts=5,
            timeout=10,
        ).and_then(
            lambda b: Ok(round(Decimal(b / 10 ** config.token_decimals.eth[token_address]), config.round_ndigits)),
        )
    else:
        return rpc.eth_get_balance(config.nodes[Network.ETH], address, proxies=config.proxies, attempts=5, timeout=10).and_then(
            lambda b: Ok(round(Decimal(b / 10**18), config.round_ndigits)),
        )


def get_token_decimals(token_address: str, config: Config) -> Result[int]:
    return erc20.get_decimals(config.nodes[Network.ETH], token_address, timeout=10, proxies=config.proxies, attempts=5)


def _get_sdai_balance(nodes: list[str], address: str, proxies: list[str], round_ndigits: int, attempts: int) -> Result[Decimal]:
    data = abi.encode_function_input_by_signature("maxWithdraw(address)", [address])
    res = rpc.eth_call(nodes, EthTokenAddress.SDAI, data, proxies=proxies, attempts=attempts)
    if isinstance(res, Err):
        return res
    balance = to_int(hexstr=res.ok)
    return Ok(Decimal(str(round(balance / 10**18, ndigits=round_ndigits))))



from __future__ import annotations

from enum import Enum, unique

DEFAULT_ETH_NODES = ["https://ethereum.publicnode.com", "https://rpc.ankr.com/eth"]


@unique
class Coin(str, Enum):
    BTC = "BTC"
    ETH = "ETH"
    USDT = "USDT"
    USDC = "USDC"
    SDAI = "SDAI"

    @classmethod
    def usd_coins(cls) -> list[Coin]:
        return [Coin.USDT, Coin.USDC, Coin.SDAI]


@unique
class Network(str, Enum):
    BTC = "btc"
    ETH = "eth"


@unique
class EthTokenAddress(str, Enum):
    USDT = "0xdac17f958d2ee523a2206206994597c13d831ec7"
    USDC = "0xa0b86991c6218b36c1d19d4a2e9eb0ce3606eb48"
    SDAI = "0x83f20f44975d03b1b09e64809b757c47f942beea"



from decimal import Decimal

from mm_std import Ok, PrintFormat, print_table

from mm_balance.balances import Balances
from mm_balance.config import Config
from mm_balance.total import Total


def print_groups(balances: Balances, config: Config, prices: dict[str, Decimal]) -> None:
    for group_index, group in enumerate(config.groups):
        group_balances = balances.get_group_balances(group_index, group.network)
        _print_group(group, group_balances, config, prices)


def _print_group(group: Config.Group, group_balances: list[Balances.Balance], config: Config, prices: dict[str, Decimal]) -> None:
    rows = []
    balance_sum = Decimal(0)
    usd_sum = Decimal(0)
    for address_task in group_balances:
        row = [address_task.address, address_task.balance.ok_or_err()]  # type: ignore[union-attr]
        if isinstance(address_task.balance, Ok):
            balance_sum += address_task.balance.ok
            if config.price:
                balance_usd = round(address_task.balance.ok * prices[group.coin], config.round_ndigits)
                usd_sum += balance_usd
                row.append(balance_usd)
        rows.append(row)

    sum_row = ["sum", round(balance_sum, config.round_ndigits)]
    if config.price:
        sum_row.append(round(usd_sum, config.round_ndigits))
    rows.append(sum_row)

    if group.sum_ratio < Decimal(1):
        sum_ratio_row = [f"sum_ratio, {group.sum_ratio}", round(balance_sum * group.sum_ratio, config.round_ndigits)]
        if config.price:
            sum_ratio_row.append(round(usd_sum * group.sum_ratio, config.round_ndigits))
        rows.append(sum_ratio_row)

    table_headers = ["address", "balance"]
    if config.price:
        table_headers += ["usd"]
    print_table(group.name, table_headers, rows)


def print_prices(config: Config, prices: dict[str, Decimal]) -> None:
    if config.price:
        rows = [[k, round(v, config.round_ndigits)] for (k, v) in prices.items()]
        print_table("price", ["coin", "usd"], rows)


def print_total(config: Config, balances: Balances, prices: dict[str, Decimal]) -> None:
    total = Total.calc(balances, prices, config)
    total.print(PrintFormat.TABLE, prices, config)
